/* SPDX-License-Identifier: MIT */
/*
 * Copyright (c) 2022 Solidigm.
 *
 * Author: leonardo.da.cunha@solidigm.com
 */

#include "telemetry-log.h"
bool solidigm_telemetry_log_header_parse(const struct telemetry_log *tl);
