// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (c) 2022 Solidigm.
 *
 * Author: leonardo.da.cunha@solidigm.com
 */

int solidigm_get_telemetry_log(int argc, char **argv, struct command *cmd, struct plugin *plugin);
