/* SPDX-License-Identifier: GPL-2.0-or-later */
/*
 * Copyright (c) 2023 Solidigm.
 *
 * Author: leonardo.da.cunha@solidigm.com
 */

#include <inttypes.h>
#include "util/json.h"
void sldgm_id_ctrl(uint8_t *vs, struct json_object *root);
